#include "srecd.h"
#include <assert.h>

int main(int ac,char **av)
{
  FILE *src;

  printf(" -= SRECD 1.0 =-\n Gdansk 2010.08.23\n\n");
  if(ac!=2)
    {
      printf(" usage: %s <filename>\n\n",av[0]);
      exit(-1);
    }
  if(src=fopen(av[1],"r"))
    {
      printf("[file %s opened...]\n\n",av[1]);
      Code=loadse(src);
      Stack=NIL;
      Dump=NIL;
      dict_init();
      T=new_sym("t");
      Env_push("t",T);
      run();
      return 0;
    }
  else err("[file input problem]");
  return -1; 
}

/****************************************************************************/

void err(char *msg)
{
  fprintf(stderr,"ERR:%s\n",msg);
  exit(-1);
}

/** Symbolic Expressions: ***************************************************/
SE *new_num(int n)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=NUM;
      refcount(e)=0;
      numval(e)=n;
    }
   else err("new_num failed");
  return e;
}

SE *new_sym(char *s)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=SYM;
      refcount(e)=0;
      symval(e)=s;
    }
   else err("new_sym failed");
  return e;
}

SE *new_str(char *s)
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=STR;
      refcount(e)=0;
      strval(e)=s;
    }
   else err("new_str failed");
  return e;
}

SE *new_cons(SE *h,SE *t) // inline??
{
  SE *e=new(SE);
  if(e)
    {
      type(e)=CONS;
      refcount(e)=0;
      head(e)=h;
      tail(e)=t;
      addref(h); // ?
      addref(t); // ?
    }
   else err("new_cons failed");
  return e;
}

int eq(SE *a,SE *b)
{
  if(a==b) return 1;
  if(a==NIL || b==NIL) return 0;
  if(type(a)!=type(b)) return 0;
  switch(type(a))
    {
    case NUM: if(numval(a)==numval(b)) return 1;
               else return 0;
    case SYM: if(strcmp(symval(a),symval(b))==0) return 1;
               else return 0;
    case STR: if(strcmp(strval(a),strval(b))==0) return 1;
               else return 0;
    default: break;
    }
  return 0;
}

void addref(SE *e) // inline?!
{
  if(e) refcount(e)++;
}

void remref(SE *e) // inline?!
{
  if(e && --(refcount(e))<=0)
    {
      if(type(e)==CONS)
      	{
      	  remref(head(e));
      	  remref(tail(e));
      	}
      free(e);
    }
}

/** Parser: *****************************************************************/
/* -- scanner: ----------------------------------------------- */
#define MAX_TOKEN 255

typedef enum{TNUM,TSYM,TSTR,TLPAR,TRPAR,TQUOTE,TEOF} toktype;

struct
{
  toktype type;
  char buf[MAX_TOKEN+1];
}token;

int (*c_in)(void);
void (*c_putback)(int);

/* -- stream input ------------------------------------------- */
FILE *f_in;
int fc_in(){return fgetc(f_in);} // getc????
void fc_putback(int c){ungetc(c,f_in);}

/* -- string input ------------------------------------------- */
char *s_in;

int sc_in()
{
  int c=*s_in;
  s_in++;
  if(c=='\0') return EOF;
  return c;
}

void sc_putback(int c){s_in--;}

/* -- scanner itself ----------------------------------------- */
void get_token()
{
  int c,i=0;
 get_token_poczatek:
  while((c=c_in())==' ' || c=='\n' || c=='\t') ;
  // omijamy komentarze, czyli bloki { ... } :
  if(c=='{')
    {
      while((c = c_in()) != '}' && c!=EOF) ;
      goto get_token_poczatek;
    }
 
  switch(c)
    {
    case EOF: token.type=TEOF; return;
    case '(': token.type=TLPAR; return;
    case ')': token.type=TRPAR; return;
    case '\'': token.type=TQUOTE; return;
    case '"':
      token.type=TSTR;
      while(i<MAX_TOKEN && (c=c_in())!=EOF && c!='"')
	token.buf[i++]=c;
      token.buf[i]='\0';
      return;
    default:
      do
	{
	  token.buf[i++]=tolower(c);
	}
      while(i<MAX_TOKEN
	    && (c=c_in())!=EOF
	    && c!=' ' && c!='\n' && c!='\t'
	    && c!='(' && c!=')' && c!='\'');
      c_putback(c);
      token.buf[i]='\0';
      if(isdigit(token.buf[0])
	 || (i>1 && token.buf[0]=='-' && isdigit(token.buf[1])))
	 token.type=TNUM;
        else
	 token.type=TSYM;
    }
}

/* -- parser ------------------------------------------------- */
SE *get_tail();

SE *get_se()
{
  switch(token.type)
    {
    case TEOF: return new_str(strdup("#eof"));
    case TNUM: return new_num(atoi(token.buf));
    case TSYM: return new_sym(strdup(token.buf));
    case TSTR: return new_str(strdup(token.buf));
    case TQUOTE: get_token(); /* ! */
                 return new_cons(new_sym(strdup("quote")),
                                 new_cons(get_se(),NIL));
    case TLPAR: return get_tail();
    default: break; /* err! */
    }
 return NIL; /* notreached */
}

SE *get_tail()
{
  SE *hd,*tl;
  get_token();
  if(token.type==TRPAR) return NIL;
  hd=get_se();
  tl=get_tail();
  return new_cons(hd,tl);
}

SE *read_se()
{
  get_token();
  return get_se();
}

/* -- output ------------------------------------------------- */
void (*c_out)(char *);

/* -- stream output ------------------------------------------ */
FILE *f_out;
void fc_out(char *s){fprintf(f_out,"%s",s);}

/* -- string output ------------------------------------------ */
#define MAX_BUF 1023
char s_buf[MAX_BUF+1];
int _sbufc;
void sc_out(char *s)
{
  while(_sbufc<MAX_BUF && *s!='\0') s_buf[_sbufc++]=*(s++);
  s_buf[_sbufc]='\0';
}

/* -- still output ------------------------------------------- */
void write_tail(SE *);

void write_se(SE *e)
{
  char numbuf[23];
  if(e==NIL)
    {
      c_out("()");
      return;
    }
  switch(type(e))
    {
    case NUM: sprintf(numbuf,"%d",numval(e));
              c_out(numbuf); break;
    case SYM: c_out(symval(e)); break;
    case STR: c_out("\""); c_out(strval(e));c_out("\""); break;
    case CONS: c_out("("); write_tail(e); break;
    }
}

void write_tail(SE *e)
{
  if(e==NIL){c_out(")"); return;}
  write_se(head(e));
  if(tail(e)==NIL){c_out(")"); return;}
  if(isatom(tail(e)))
    {
      c_out(" . ");
      write_se(tail(e));
      c_out(")");
      return;
    }
  c_out(" ");
  write_tail(tail(e));
}

/* -- TADAAM ------------------------------------------------- */
void display(SE *e)
{
  f_out=stdout;
  c_out=fc_out;
  write_se(e);
}

SE *read()
{
  f_in=stdin;
  c_in=fc_in;
  c_putback=fc_putback;
  return read_se();
}

void savese(FILE *f,SE *e)
{
  f_out=f;
  c_out=fc_out;
  write_se(e);
}

SE *loadse(FILE *f)
{
  f_in=f;
  c_in=fc_in;
  c_putback=fc_putback;
  return read_se();
}

char *se2str(SE *e)
{
  c_out=sc_out;
  _sbufc=0;
  write_se(e);
  return strdup(s_buf);
}

SE *parse(char *s)
{
  s_in=s;
  c_in=sc_in;
  c_putback=sc_putback;
  return read_se();
}

/** The dictionary implementation: ******************************************/
void dict_init()
{
  int i;
  for(i=0;i<TABLESIZE;i++)
    dictionary[i]=NIL;
}

int hashfun(char *s)
{
  unsigned char hash = 0; 
  while(*s!='\0')
  {
    int c = *s;
    hash = ((hash << 5) - hash) + c;
    s++;
  }
  return (hash%TABLESIZE);
}

void Env_push(char *s,SE *e)
{
  int h=hashfun(s);
  Record *r=dictionary[h],*r2;
  while(r!=NIL && (strcmp(r->key,s)!=0)) r=r->next;
  if(r!=NIL)
    { 
      r->valstack=new_cons(e,r->valstack);
      return;
    }
  r2=new(Record);
  r2->valstack=new_cons(e,NIL);
  r2->key=s;
  r2->next=dictionary[h];
  dictionary[h]=r2;
}

SE *Env_get(char *s)
{
  int h=hashfun(s);
  Record *r=dictionary[h];
  while(r!=NIL &&strcmp(r->key,s)!=0) r=r->next;
  if(r==NIL || r->valstack==NIL) return NIL;
  return head(r->valstack);
}

void Env_pop(char *s)
{
  int h=hashfun(s);
  Record *r=dictionary[h],*r2;
  while(r!=NIL &&strcmp(r->key,s)!=0) r=r->next;
  assert(r!=NIL);
  if(r->valstack!=NIL)
    {
      SE *gc;
      gc=r->valstack;
      r->valstack=tail(r->valstack);
      addref(r->valstack);
      remref(gc);
    }
  else
    {
      /* remove record... */
      if(dictionary[h]==r)
	{
	  dictionary[h]=dictionary[h]->next;
	}
      else
	{
	  for(r2=r;r2!=NIL && r2->next!=r;r2=r2->next) ;
	  assert(r2!=NIL);
	  r2->next=r->next;
	}
      free(r);
    }
}

/** and SRECD machine implementation: ***************************************/
void run()
{
  SE *gc;

 top:

  /*
  printf(".s: ");display(Stack); printf("\n");
  printf(".c: ");display(Code); printf("\n");
  printf(".d: ");display(Dump); printf("\n");
  */

  if(Code==NIL) stop();
  switch(numval(head(Code)))
    {
    case oNIL: gc=Stack;
               Stack=new_cons(NIL,Stack);
	       addref(Stack);
	       remref(gc);
               gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;

    case oLU: gc=Stack;
              Stack=new_cons(Env_get(strval(head(tail(Code)))),Stack);
	      addref(Stack);
	      remref(gc);
              gc=Code;
              Code=tail(tail(Code));
	      addref(Code);
	      remref(gc);
	      goto top;

    case oLDC: gc=Stack;
               Stack=new_cons(head(tail(Code)),Stack);
	       addref(Stack);
	       remref(gc);
               gc=Code;
               Code=tail(tail(Code));
	       addref(Code);
	       remref(gc);
	       goto top;

    case oBIND: Env_push(strval(head(tail(Code))),head(Stack));
                gc=Stack;
                Stack=tail(Stack);
		addref(Stack);
		remref(gc);
                gc=Code;
                Code=tail(tail(Code));
 	        addref(Code);
	        remref(gc);
	        goto top;

    case oUNBIND: Env_pop(strval(head(tail(Code))));
                  gc=Code;
                  Code=tail(tail(Code));
	          addref(Code);
	          remref(gc);
	          goto top;

    case oEQ: gc=Stack; 
              if(eq(head(tail(Stack)), head(Stack))==1)
	        Stack=new_cons(T,tail(tail(Stack)));
               else Stack=new_cons(NIL,tail(tail(Stack)));
              addref(Stack);
	      remref(gc);
              gc=Code;
              Code=tail(Code);
	      addref(Code);
	      remref(gc);
	      goto top;

    case oNUM: gc=Stack;
               if(isnum(head(Stack))) Stack=new_cons(T, tail(Stack));
                 else Stack=new_cons(NIL, tail(Stack));
	       addref(Stack);
	       remref(gc);
               gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;

    case oATOM: gc=Stack;
                if(isatom(head(Stack))) Stack=new_cons(T, tail(Stack));
                  else Stack=new_cons(NIL, tail(Stack));
	        addref(Stack);
 	        remref(gc);		
                gc=Code;
                Code=tail(Code);
		addref(Code);
	        remref(gc);
	        goto top;

    case oCAR: gc=Stack;
               Stack=new_cons(head(head(Stack)), tail(Stack));
	       addref(Stack);
	       remref(gc);
               gc=Code;
               Code=tail(Code);
               addref(Code);
	       remref(gc);
	       goto top;

    case oCDR: gc=Stack;
               Stack=new_cons(tail(head(Stack)), tail(Stack));
	       addref(Stack);
	       remref(gc);
               gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;

    case oCONS: gc=Stack;
                Stack=new_cons(new_cons(head(Stack),head(tail(Stack))),tail(tail(Stack)));		
 	        addref(Stack);
	        remref(gc);
                gc=Code;
                Code=tail(Code);
		addref(Code);
	        remref(gc);
	        goto top;

    case oSEL: gc=Dump;
               Dump=new_cons(tail(tail(tail(Code))),Dump);
	       addref(Dump);
	       remref(gc);
	       gc=Code;
               if(head(Stack)!=NIL)
		 Code=head(tail(Code));
	       else
	         Code=head(tail(tail(Code)));
	       addref(Code);
	       remref(gc);
	       gc=Stack;
	       Stack=tail(Stack);
	       addref(Stack);
	       remref(gc);
	       goto top;

    case oSER: gc=Code;
               if(head(Stack)!=NIL)
		 Code=head(tail(Code));
	       else
	         Code=head(tail(tail(Code)));
	       addref(Code);
	       remref(gc);
	       gc=Stack;
	       Stack=tail(Stack);
	       addref(Stack);
	       remref(gc);
	       goto top;

    case oAP: gc=Dump;
              Dump=new_cons(tail(Code),Dump);
	      addref(Dump);
	      remref(gc);
	      gc=Code;
              Code=head(Stack);
	      addref(Code);
	      remref(gc);
	      gc=Stack;
              Stack=tail(Stack);
	      addref(Stack);
	      remref(gc);
	      goto top;

    case oTRAP: gc=Code;
                Code=head(Stack);
		addref(Code);
		remref(gc);
		gc=Stack;
                Stack=tail(Stack);
		addref(Stack);
		remref(gc);
 	        goto top;

    case oRTN: gc=Code;
               Code=head(Dump);
	       addref(Code);
	       remref(gc);
	       gc=Dump;
               Dump=tail(Dump);
	       addref(Dump);
	       remref(gc);
	       goto top;

    case oDISP: display(head(Stack));
                printf("\n"); // 
		gc=Code;
                Code=tail(Code);
		addref(Code);
		remref(gc);
		goto top;

    case oREAD: printf(">"); // 
                gc=Stack;
                Stack=new_cons(read(),Stack);
		addref(Stack);
		remref(gc);
		gc=Code;
                Code=tail(Code);
		addref(Code);
		remref(gc);
		goto top;
	       
    case oADD: gc=Stack;
               Stack=new_cons(new_num(numval(head(Stack))+numval(head(tail(Stack)))),tail(tail(Stack)));
	       addref(Stack);
	       remref(gc);
	       gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;

    case oMUL: gc=Stack;
               Stack=new_cons(new_num(numval(head(Stack))*numval(head(tail(Stack)))),tail(tail(Stack)));
	       addref(Stack);
	       remref(gc);
	       gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;
               
    case oSUB: gc=Stack;
               Stack=new_cons(new_num(numval(head(tail(Stack)))-numval(head(Stack))),tail(tail(Stack)));
	       addref(Stack);
	       remref(gc);
	       gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;

    case oDIV: gc=Stack;
               Stack=new_cons(new_num(numval(head(tail(Stack)))/numval(head(Stack))),tail(tail(Stack)));
	       addref(Stack);
	       remref(gc);
	       gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;

    case oMOD: gc=Stack;
               Stack=new_cons(new_num(numval(head(Stack))%numval(head(tail(Stack)))),tail(tail(Stack)));
	       addref(Stack);
	       remref(gc);
	       gc=Code;
               Code=tail(Code);
	       addref(Code);
	       remref(gc);
	       goto top;

    case oGT: gc=Stack;
              if(numval(head(Stack))<numval(head(tail(Stack))))
                Stack=new_cons(T,tail(tail(Stack)));
	      else
               Stack=new_cons(NIL,tail(tail(Stack)));
	      addref(Stack);
	      remref(gc);
	      gc=Code;
              Code=tail(Code);
	      addref(Code);
	      remref(gc);
	      goto top;

    case oLT: gc=Stack;
              if(numval(head(Stack))>numval(head(tail(Stack))))
                Stack=new_cons(T,tail(tail(Stack)));
	      else
                Stack=new_cons(NIL,tail(tail(Stack)));
	      addref(Stack);
	      remref(gc);
	      gc=Code;
              Code=tail(Code);
	      addref(Code);
	      remref(gc);
  	      goto top;

    default: stop();
    }
}

void stop()
{
  display(head(Stack));
  printf("\n");
  exit(0);
}

/****************************************************************************/
