/**********************************
 * maszyna SRECD dla jezyka drcz0 *
 * S.Dercz                        *
 * Gdansk, 2010.08.08-23          *
 **********************************/

#include<stdio.h>
#include<stdlib.h>
#include<ctype.h>
#include<string.h>

#define new(TYPE) (TYPE *)malloc(sizeof(TYPE))

void err(char *);

/** Symbolic Expressions: ***************************************************/
typedef struct SE SE;
typedef enum{NUM,SYM,STR,CONS}setype;

struct SE
{
  setype type;
  int ref_count;

  union
  {
    int num;
    char *str;
    SE *cons[2];
  };
};

#define NIL NULL
SE *T;

#define type(S) (S)->type
#define refcount(S) (S)->ref_count
#define numval(S) (S)->num
#define symval(S) (S)->str
#define strval(S) (S)->str
#define head(S) (S)->cons[0]
#define tail(S) (S)->cons[1]
#define isatom(S) ((S)==NIL||(S)->type!=CONS)
#define isnum(S) ((S)->type==NUM)
#define isstr(S) ((S)->type==STR)

SE *new_num(int);
SE *new_sym(char *);
SE *new_str(char *);
SE *new_cons(SE *,SE *); // inline?!

int eq(SE *,SE *);

void addref(SE *); // inline?!
void remref(SE *); // inline?!


/** Parser: *****************************************************************/
void display(SE *);
SE *read();

void savese(FILE *,SE *);
SE *loadse(FILE *);

char *se2str(SE *);
SE *parse(char *);


/** SRECD machine: **********************************************************/
SE *Stack,*Code,*Dump;
void Env_push(char *,SE *);
SE *Env_get(char *);
void Env_pop(char *);

void run();
void stop();

enum _opcodes
{
  oNIL,
  oLU,
  oLDC,
  oBIND,
  oUNBIND,
  oEQ,
  oNUM,
  oATOM,
  oCAR,
  oCDR,
  oCONS,
  oSEL,
  oSER,
  oAP,
  oTRAP,
  oRTN,
  oDISP,
  oREAD,
  oADD,
  oMUL,
  oSUB,
  oDIV,
  oMOD,
  oLT,
  oGT
};

/** A dictionary: ***********************************************************/
typedef struct _record Record;

struct _record
{
  char *key;
  SE *valstack;
  Record *next;
}; 

#define TABLESIZE 1019 /* :) */
Record *dictionary[TABLESIZE];

void dict_init();
int hashfun(char *);
